/* $Id: linuxsuid.h,v 1.2 1998/11/17 21:52:09 becka Exp $
******************************************************************************

   KGI - Linux specific parts

   Copyright (C) 1998	Andreas Beck	[becka@ggi-project.org]
   Copyright (C) 1997	Steffen Seeger	[seeger@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#ifndef _KGI_SYS_LINUXSUID_H
#define _KGI_SYS_LINUXSUID_H

#define MAX_NR_CONSOLES 1

typedef void * mem_vaddr;      /* the virtual address type     */

#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/perm.h>
#include <sys/mman.h>

#include <asm/io.h>

#define	IS_CRITICAL	unsigned long __critical_flags 
#define	ENTER_CRITICAL	(__critical_flags++)
#define	LEAVE_CRITICAL	(__critical_flags--)

/*	use these macros to allocate memory. 
	kgi_malloc_small() may be used in IRQ environments, but may
	likely fail for big chunks requested there. In general it shall
	be used when allocating small (< PAGESIZE) chunks of memory.
	kgi_malloc/free_huge() are used to allocate big chunks of memory
	such as console backround buffers.
*/

#define kgi_malloc_small(size)	malloc(size)
#define	kgi_free_small(addr)	free(addr)

#define	kgi_malloc_huge(size)	malloc(size)
#define	kgi_free_huge(addr)	free(addr)

#define NOT_TESTED \
	printf(\
    	"\nGood! You hit untested code in %s line %i. Please report whether\n"\
	"there were any difficulties or not. (Simply send this message and a\n"\
	"short explanation of your configuration to " MAINTAINER ".\n\n\n",\
	__FILE__,__LINE__)

/* Defines for struct kgi_mmio_region */
#define	MMIO_RD		0x0001
#define	MMIO_WR		0x0002 
#define	MMIO_RDWR	(MMIO_RD | MMIO_WR)
#define	MMIO_SHARED	0x0008

extern inline int suser(void)
{
	return 1;       /* We always run with super user priviledges. Maybe we
			   should distinguish real root/suid root here ? */
}
extern inline void * phys_to_virt(unsigned long address)
{
	NOTICE("ARGH someone called phys_to_virt !\n");
	return (void *) -(1<<10);  /* cause a segfault */
	/*      return (void *)address; */
}
#define phys_to_virt You_shall_not_use_phys_to_virt()

/*	A ..._region is used to describe a region in a physical address space.
**	Per physical address space, there are the following actions possible
**	with a region:
**
**	int ..._check_region(struct ..._region *)
**		Before doing *any* access to a region, you have to check if the
**		region is not served by another driver. check_region() returns
**		EOK (== 0) if noone else claimed (a part) of this region before.
**		or a non-zero value otherwise.
**
**	..._vaddr ..._claim_region(struct ..._region *)
**		This makes a claim that you are going to handle this region
**		from now on. You must have checked a region successfully
**		before claiming it. A non-zero virtual address should be
**		returned.
**
**	..._vaddr ..._free_region(struct ..._region *)
**		Release a claim made earlier on this region. You must not
**		access the region after releasing your claim.
**
**	..._vaddr ..._alloc_region(struct ..._region *r)
**		Allocate and claim a free region of size r->size with r->decode
**		being the address lines decoded. This returns a non-zero
**		virtual address if successful or a zero value otherwise.
**
**	void ..._out<width>(?int<width> val, ..._vaddr vaddr)
**		Do the neccessary actions to write the value 'val' at
**		address 'addr'. 'addr' must be aligned to the <width> of
**		the value being written. The write cycle generated is of width
**		<width>.
**
**	kgi_uint<width> ..._in<width>(..._vaddr vaddr)
**		Do the neccessary actions to reat the cell at
**		address 'addr' which must be alligned to a cell width.
**		The read cycle generated is of width <width>.
*/

/*	Per region, we have a physical and a virtual address space and
**	certain information is stored with each region. As this is
**	common to all regions, let's make life easier.
*/

/*	Intel style in/out I/O space
**
**	The io_in.. / io_out.. functions generate the neccesary hardware
**	actions	to do a read/write cycle in a cards I/O space. (Registers
**	accessed via the in.. / out.. instructions on i386 architecture.)
**
**	The function names encode direction (in == read cycle, out == write
**	cycle) and register width, which is 8, 16 or 32 bits. Any address given
**	__must__ be aligned to register widths, otherwise results may
**	be undefined!
*/

typedef kgiu16 io_addr; 	/* the physical address type	*/
typedef kgiu16 io_vaddr; 	/* the virtual address type	*/

struct io_region 
{
	io_vaddr baseptr;	/* virtual base address		*/
	io_addr base;		/* physical base address	*/
	io_addr size;		/* size of region		*/
	io_addr decode;		/* decoded phys address lines	*/
	char *name;		/* name of the region		*/
};

#define	IO_NULL		((io_vaddr) 0)

#ifdef IO_CHECK_REGION_ACTUALLY_WORKS
/* This will currently always fail */
extern inline int io_check_region(struct io_region *r)
{
	register kgi_sint foo = check_region(r->base, r->size);

	DEBUG2("checking io_region %s, base = %.4x, size %i, is %s", \
		r->name, r->base, r->size, foo ? "not free" : "free");
	return foo;
}
#else
#define io_check_region(x)	0
#endif /* IO_CHECK_REGION_ACTUALLY_WORKS */

extern inline io_vaddr io_claim_region(struct io_region *r)
{
	PASSERT(r->baseptr == IO_NULL, "claim of claimed io_region '%s'!",\
		r->name);

	if (r->base+r->size<=0x400) {
		if (ioperm(r->base,r->size,1)) 
			perror("io_claim_region-ioperm");
	} else {
		if (iopl(2)) perror("io_claim_region-iopl");
	}
	r->baseptr = r->base;

	DEBUG2("claiming io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);
	return r->baseptr;
}

extern inline io_vaddr io_free_region(struct io_region *r)
{
	DEBUG2("freeing io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);

	if (r->base+r->size<=0x400) {
		if (ioperm(r->base,r->size,0)) 
			perror("io_free_region-ioperm");
	} else {
		DEBUG2("iopl _NOT_ reversed to 0.\n");
	}
	return (r->baseptr = IO_NULL);
}

extern inline io_vaddr io_alloc_region(struct io_region *r)
{
	/*	nothing appropriate under Linux. AFAIK - seeger
	*/
	return IO_NULL;
}

extern inline kgiu8 io_in8(io_vaddr addr) 
{
	return inb(addr);
}

extern inline kgiu16 io_in16(io_vaddr addr)
{
	return inw(addr);
}

extern inline kgiu32 io_in32(io_vaddr addr)
{
	return inl(addr);
}

extern inline void io_ins8(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insb(addr, buf, cnt);
}

extern inline void io_ins16(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insw(addr, buf, cnt);
}

extern inline void io_ins32(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insl(addr, buf, cnt);
}

extern inline void io_out8(kgiu8 val, io_vaddr addr)
{
	outb(val, addr);
}

extern inline void io_out16(kgiu16 val, io_vaddr addr)
{
	outw(val, addr);
}

extern inline void io_out32(kgiu32 val, io_vaddr addr)
{
	outl(val, addr);
}

extern inline void io_outs8(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsb(addr, buf, cnt);
}

extern inline void io_outs16(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsw(addr, buf, cnt);
}

extern inline void io_outs32(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsl(addr, buf, cnt);
}

/*	memory I/O space
**
**	This is the 'normal' shared memory I/O space accessed using the
**	mov instructions on i386 architecture.
*/

typedef kgi_addr mem_addr; 	/* the physical address type	*/

struct mem_region 
{
	mem_vaddr baseptr;	/* virtual base address		*/
	mem_addr base;		/* physical base address	*/
	mem_addr size;		/* size of region		*/
	mem_addr decode;	/* decoded phys address lines	*/
	char *name;		/* name of the region		*/
};

#define	MEM_NULL	((mem_vaddr) 0)

extern int _suidtarget_dev_mem;

extern inline kgi_sint mem_check_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
	*/
	DEBUG2("checking mem_region %s, base %.8x, size %i, is always free!",\
		r->name, r->base, r->size);
	return EOK;
}

extern inline mem_vaddr mem_claim_region(struct mem_region *r)
{
	PASSERT(r->baseptr == MEM_NULL, "claim of claimed mem_region '%s'!", \
		r->name);

	r->baseptr = mmap(NULL,r->size,PROT_READ|PROT_WRITE,MAP_SHARED,_suidtarget_dev_mem,r->base);

	DEBUG2("claiming mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return r->baseptr;
}

extern inline mem_vaddr mem_free_region(struct mem_region *r)
{
	PASSERT(r->baseptr, "free of freed region '%s'!", r->name);

	munmap(r->baseptr,r->size);

	DEBUG2("freeing mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return (r->baseptr = NULL);
}

extern inline mem_vaddr mem_alloc_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
 	*/
	return MEM_NULL;
}

extern inline kgiu8 mem_in8(mem_vaddr addr)
{
	return  *((kgiu8 *)addr);
}

extern inline kgiu16 mem_in16(mem_vaddr addr)
{
	return  *((kgiu16 *)addr);
}

extern inline kgiu32 mem_in32(mem_vaddr addr)
{
	return  *((kgiu32 *)addr);
}

extern inline void mem_out8(kgiu8 val, mem_vaddr addr)
{
	*((kgiu8 *)addr)=val;
}

extern inline void mem_out16(kgiu16 val, mem_vaddr addr)
{
	*((kgiu16 *)addr)=val;
}

extern inline void mem_out32(kgiu32 val, mem_vaddr addr)
{
	*((kgiu32 *)addr)=val;
}

#define	MEM_INS(x) \
extern inline void mem_ins##x(mem_vaddr addr, kgiu##x *buf, kgi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		*(buf++) = mem_in##x(addr);			\
	}							\
}

#define	MEM_OUTS(x) \
extern inline void mem_outs##x(mem_vaddr addr, kgiu##x *buf, kgi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		mem_out##x(*(buf++), addr);			\
	}							\
}

MEM_INS(8)
MEM_INS(16)
MEM_INS(32)

MEM_OUTS(8)
MEM_OUTS(16)
MEM_OUTS(32)

#undef	MEM_OUTS
#undef	MEM_INS

/*	PCI configuration space
**
**	These functions should be used to access the PCI configuration space.
**	As the address is given by the hardware wiring, a registration
**	scheme doesn't make sense. But I/O is pretty useful :)
*/
#include <linux/version.h>
#define LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,90)
	#include <linux/pci.h>
	#include <linux/bios32.h>
#else
	#define __KERNEL__
	#include <linux/pci.h>
	#undef __KERNEL__
#endif

typedef kgi_addr pcicfg_vaddr;
#define	PCICFG_NULL	((pcicfg_vaddr) -1)
#define	PCICFG_VADDR(bus, slot, fn)	\
	((pcicfg_vaddr)((((bus) & 0xFF) << 24) | (PCI_DEVFN(slot,fn) << 16)))

#define	PCIARGS	(addr >> 24) & 0xFF, (addr >> 16) & 0xFF, addr & 0xFF

#if CONFIG_PCI

#define CONFIG_CMD(bus, device_fn, where)   (0x80000000 | (bus << 16) | (device_fn << 8) | (where & ~3))

/*
 * Functions for accessing PCI configuration space with type 1 accesses
 */

extern inline int pcibios_read_config_byte(unsigned char bus, unsigned char device_fn,
				unsigned char where, unsigned char *value)
{
	IS_CRITICAL;
	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);
	*value = inb(0xCFC + (where&3));
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;
}

extern inline int pcibios_read_config_word (unsigned char bus,
	unsigned char device_fn, unsigned char where, unsigned short *value)
{
	IS_CRITICAL;

	if (where&1) return PCIBIOS_BAD_REGISTER_NUMBER;

	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);    
	*value = inw(0xCFC + (where&2));
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;    
}

extern inline int pcibios_read_config_dword (unsigned char bus, unsigned char device_fn, 
				unsigned char where, unsigned int *value)
{
	IS_CRITICAL;

	if (where&3) return PCIBIOS_BAD_REGISTER_NUMBER;

	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);
	*value = inl(0xCFC);
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;    
}

extern inline int pcibios_write_config_byte (unsigned char bus, unsigned char device_fn, 
				 unsigned char where, unsigned char value)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);    
	outb(value, 0xCFC + (where&3));
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;
}

extern inline int pcibios_write_config_word (unsigned char bus, unsigned char device_fn, 
				 unsigned char where, unsigned short value)
{
	IS_CRITICAL;

	if (where&1) return PCIBIOS_BAD_REGISTER_NUMBER;

	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);
	outw(value, 0xCFC + (where&2));
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;
}

extern inline int pcibios_write_config_dword (unsigned char bus, unsigned char device_fn, 
				  unsigned char where, unsigned int value)
{
	IS_CRITICAL;

	if (where&3) return PCIBIOS_BAD_REGISTER_NUMBER;

	ENTER_CRITICAL;
	outl(CONFIG_CMD(bus,device_fn,where), 0xCF8);
	outl(value, 0xCFC);
	LEAVE_CRITICAL;
	return PCIBIOS_SUCCESSFUL;
}

extern inline kgiu8 pcicfg_in8(pcicfg_vaddr addr)
{
	kgiu8 tmp;
	pcibios_read_config_byte(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.2x\n", addr, tmp);
	return tmp;
}

extern inline kgiu16 pcicfg_in16(pcicfg_vaddr addr)
{
	kgiu16 tmp;
	pcibios_read_config_word(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.4x\n", addr,  tmp);
	return tmp;
}

extern inline kgiu32 pcicfg_in32(pcicfg_vaddr addr)
{
	kgiu32 tmp;
	pcibios_read_config_dword(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.8x\n", addr, tmp);
	return tmp;
}

extern inline void pcicfg_out8(kgiu8 val, pcicfg_vaddr addr)
{
	pcibios_write_config_byte(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.2x\n", addr, val);
}

extern inline void pcicfg_out16(kgiu16 val, pcicfg_vaddr addr)
{
	pcibios_write_config_word(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.4x\n", addr, val);
}

extern inline void pcicfg_out32(kgiu32 val, pcicfg_vaddr addr)
{
	pcibios_write_config_dword(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.8x\n", addr, val);
}

#undef PCIARGS

#else	/* #if CONFIG_PCI */

extern inline kgiu8 pcicfg_in8(pcicfg_vaddr addr)
{
	return -1;
}

extern inline kgiu16 pcicfg_in16(pcicfg_vaddr addr)
{
	return -1;
}

extern inline kgiu32 pcicfg_in32(pcicfg_vaddr addr)
{
	return -1;
}

extern inline void pcicfg_out8(kgiu8 val, pcicfg_vaddr addr)
{
}

extern inline void pcicfg_out16(kgiu16 val, pcicfg_vaddr addr)
{
}

extern inline void pcicfg_out32(kgiu32 val, pcicfg_vaddr addr)
{
}

#endif /* #if CONFIG_PCI */

#endif /* _KGI_SYS_LINUXSUID_H */
